# Copyright 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

"""Redis connection"""
import operator
import socket
import urlparse
from functools import partial, wraps
from urlparse import urlsplit

import redis
from redis.client import dict_merge
from redis import StrictRedis
from redis.sentinel import Sentinel

LOCAL_DOMAIN = socket.gethostname().split('.', 1)[1]
DEFAULT_REDIS_GROUP = 'cumulus-netq'
DEFAULT_REDIS_MASTER_PORT = 6379
DEFAULT_REDIS_DB = 0
DEFAULT_MASTER_CONNECTION = "redis://ccc.{domain}:{port}/{db}".format(
    domain=LOCAL_DOMAIN, port=DEFAULT_REDIS_MASTER_PORT, db=DEFAULT_REDIS_DB)
DEFAULT_REDIS_KWARGS = dict(socket_timeout=5)


class ConnectionError(Exception):
    """NetQ Redis connection Error"""
    pass


class ConfigError(Exception):
    """NetQ Config file Error"""
    pass


def skip_if_down(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except redis.ConnectionError:
            pass

    return wrapper


def _from_url(url, redis_class=StrictRedis, **kwargs):
    connection_kwargs = dict(DEFAULT_REDIS_KWARGS)
    connection_kwargs.update(kwargs)
    return redis_class.from_url(url, **connection_kwargs)


def host_port_from_url(url):
    sp = urlsplit(url)
    return sp.hostname, sp.port


@skip_if_down
def is_role(url, role):
    """Is specific role

    Given Redis URL instance, figure out if it will resolve to given role.
    param: url: Redis instance URL
    param: role: role string ('master', 'sentinel')
    """
    info = _from_url(url).info()
    # TODO Fix this logic
    if 'role' in info:
        return info['role'] == role
    return info['redis_mode'] == role


is_master = partial(is_role, role='master')  # pylint: disable=locally-disabled, invalid-name
is_sentinel = partial(is_role, role='sentinel')  # pylint: disable=locally-disabled, invalid-name


def normalize_redis_url(url):
    """Normalize Redis url

    Normalize url to have full form of Redis URL
    (https://www.iana.org/assignments/uri-schemes/prov/redis)

    Input url string can be in 2 forms:
    - partial Redis URL
    - <host> [port <value>] [db <value>] ....

    params: url: raw url string
    """
    orig_url = url

    # Parse key-value format
    if ' ' in url:
        host, url = url.split(' ', 1)
        url_kw = {'host': host, 'password': '', 'db': DEFAULT_REDIS_DB}
        while url:
            vals = url.split(' ', 2)
            if len(vals) < 2:
                raise ConfigError("Invalid URL string '{}'.".format(orig_url))
            key, value = vals[0], vals[1]
            url = vals[2] if len(vals) == 3 else None
            if key in ('port', 'db', 'password', 'group'):
                url_kw[key] = value
            else:
                raise ConfigError("Invalid keyword '{}'".format(key))
        url = "redis://{password}{host}:{port}/{db}".format(**url_kw)
        if 'group' in url_kw:
            url = "{}?{}".format(url, url_kw['group'])
        return url

    # Parse url format
    if '://' not in orig_url:
        # Prepend '://' if missing for urlparse
        # to work properly
        url = 'redis://' + url

    scheme, netloc, path, query, fragment = sp = urlparse.urlsplit(url)
    if not sp.port:
        netloc = "{}:{}".format(sp.netloc, DEFAULT_REDIS_MASTER_PORT)
    if not sp.path:
        path = str(DEFAULT_REDIS_DB)
    return urlparse.urlunsplit((scheme, netloc, path, query, fragment))


def _common_query_param(seq, param):
    """Get common values for query param

    param: seq: sequence of URL to inspect
    param: param: query param name
    returns: Value(s) of a common param
    """
    seq = map(operator.attrgetter('query'), map(urlparse.urlparse, seq))
    seq = map(lambda x: urlparse.parse_qs(x).get(param), seq)
    seq = map(operator.itemgetter(0), seq)
    return set(seq)


_common_group = partial(_common_query_param, param='group')


def get_redis_master(*urls, **connection_kwargs):
    """Figure out redis master given urls.

    :param ulrs: seqence of conncetion urls
    :returns StrictRedis instance

    Urls can be in a simple domain format (i.e. redis.localhost:1234) or
    redis url format (i.e. redis://redis.localhost:1234)
    """
    connection_kwargs = dict_merge(DEFAULT_REDIS_KWARGS, connection_kwargs or
                                   {})
    redis_urls = {normalize_redis_url(url) for url in urls if url}

    if not redis_urls:
        return _from_url(DEFAULT_MASTER_CONNECTION)

    masters = filter(is_master, redis_urls)
    sentinels = filter(is_sentinel, redis_urls)

    if not sentinels:
        # No sentinels
        if not masters:
            raise ConnectionError(
                "No available masters or sentinels to connect to.")
        if len(masters) > 1:
            raise ConfigError("Multi-master case is not supported.")
        # Single master
        return _from_url(masters[0])
    else:
        if masters:
            raise ConfigError("Mixing masters and slaves is not allowed.")

        sentinel_hosts = map(host_port_from_url, sentinels)
        sentinel = Sentinel(sentinel_hosts, **connection_kwargs)
        # TODO get common group
        # redis_group = _common_group(sentinels) or DEFAULT_REDIS_GROUP
        redis_group = DEFAULT_REDIS_GROUP
        master_for_group = sentinel.master_for(redis_group)
        return master_for_group

    return None
